import{M as d,N as f,a as o,O as l,Q as v}from"./index.52551726.js";const S=()=>{};function c(){const i=a(o(S));return i.executed=l(!1),i}function a(i){return v(i)}function x(...[i,s,b={}]){const e=d(c());return f(s,(n,p)=>{var u,r;if(n===null&&p===void 0)return;if(e.isSubscribed()&&e.unsubscribe(),n===null){Object.assign(e,c());return}const t=a(o(i,s,b));(r=(u=t.response).value)!=null||(u.value=e.response),t.executed.value=t.executed.value||e.executed,Object.assign(e,t)},{immediate:!0}),e}export{x as u};
