import{d as b,c as A,o as i,i as _,k as t,w as n,F as x,aZ as C,g as a,a_ as w,a$ as m,b0 as $,b1 as h,b2 as B,ao as f,O,b3 as M,e as V,b4 as E,r,f as F,b5 as R,b6 as j,b7 as D}from"./index.52551726.js";const L={class:"settings-block"},N=b({__name:"SettingsCodeBlock",props:{engineSettings:null},setup(c){const e=c,s=A(()=>Object.entries(e.engineSettings));return(u,d)=>(i(),_("div",L,[t(a(w),{multiline:""},{default:n(()=>[(i(!0),_(x,null,C(a(s),(l,o)=>(i(),_("div",{key:o,class:"settings-block--code-line"},m(l[0])+": "+m(l[1]),1))),128))]),_:1})]))}});class P extends h(B){}const v=$(new P);const Z=b({__name:"Settings",async setup(c){let e,s;const u=[{text:"Settings"}],d=([e,s]=f(()=>v.getSettings()),e=await e,s(),e),l=([e,s]=f(()=>v.getVersion()),e=await e,s(),e),o=O(M());return V("Settings"),E(()=>D(o.value)),(T,g)=>{const S=r("p-key-value"),p=r("p-label"),k=r("p-layout-default");return i(),F(k,{class:"settings"},{header:n(()=>[t(a(R),{crumbs:u},{actions:n(()=>[t(S,{label:"Version",value:a(l),alternate:""},null,8,["value"])]),_:1})]),default:n(()=>[t(p,{label:"Color Mode",class:"settings__color-mode"},{default:n(()=>[t(a(j),{selected:o.value,"onUpdate:selected":g[0]||(g[0]=y=>o.value=y)},null,8,["selected"])]),_:1}),t(p,{label:"Orion Settings"},{default:n(()=>[t(N,{class:"settings__code-block","engine-settings":a(d)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{Z as default};
