import{d as _,a as p,c as a,e as d,r as k,o as s,f as c,w as l,k as f,g as e,az as i,i as b,aA as B,aB as h,F as y,h as D,aC as g}from"./index.52551726.js";const v=_({__name:"Blocks",setup(x){const o=p(g.getBlockDocuments),n=a(()=>{var t;return(t=o.response)!=null?t:[]}),u=a(()=>o.executed&&n.value.length==0),r=a(()=>o.executed);return d("Blocks"),(t,C)=>{const m=k("p-layout-default");return s(),c(m,{class:"blocks"},{header:l(()=>[f(e(i))]),default:l(()=>[e(r)?(s(),b(y,{key:0},[e(u)?(s(),c(e(B),{key:0})):(s(),c(e(h),{key:1,"block-documents":e(n),onDelete:e(o).refresh},null,8,["block-documents","onDelete"]))],64)):D("",!0)]),_:1})}}});export{v as default};
