import{d as u,e as _,r as l,o as p,f as m,w as o,k as e,g as n,aW as d,aX as h,aV as C,a5 as i,t as s,a4 as w}from"./index.52551726.js";const g=u({__name:"NotificationCreate",setup(y){async function c(a){try{await C.createNotification(a),i.push(s.notifications())}catch(t){w("Error creating notification","error"),console.warn(t)}}function r(){i.push(s.notifications())}return _("Create Notification"),(a,t)=>{const f=l("p-layout-default");return p(),m(f,{class:"notification-create"},{header:o(()=>[e(n(d))]),default:o(()=>[e(n(h),{action:"Create",onSubmit:c,onCancel:r})]),_:1})}}});export{g as default};
