import{d as p,c as t,e as i,r as k,g as o,o as m,f as y,w as l,k as s,aG as _,aH as b,h as f,aF as d}from"./index.52551726.js";import{u as T}from"./useRouteParam.4bc8e74f.js";import{u as g}from"./useSubscriptionWithDependencies.769857cf.js";const x=p({__name:"BlocksCatalogView",setup(v){const a=T("blockTypeSlug"),c=t(()=>a.value?[a.value]:null),n=g(d.getBlockTypeBySlug,c),e=t(()=>n.response),u=t(()=>e.value?`Block Type: ${e.value.name}`:null);return i(u),(B,S)=>{const r=k("p-layout-default");return o(e)?(m(),y(r,{key:0,class:"blocks-catalog-view"},{header:l(()=>[s(o(_),{"block-type":o(e)},null,8,["block-type"])]),default:l(()=>[s(o(b),{"block-type":o(e)},null,8,["block-type"])]),_:1})):f("",!0)}}});export{x as default};
