import{d as m,a as d,c as n,e as _,r as k,o as a,f as i,w as r,k as y,g as e,aS as g,i as x,aT as N,aU as $,F as v,h as C,aV as S}from"./index.52551726.js";const V=m({__name:"Notifications",setup(h){const t=d(S.getNotifications),c=n(()=>{var s;return(s=t.response)!=null?s:[]}),f=n(()=>t.executed&&c.value.length===0),u=n(()=>t.executed);return _("Notifications"),(s,o)=>{const l=k("p-layout-default");return a(),i(l,{class:"notifications"},{header:r(()=>[y(e(g))]),default:r(()=>[e(u)?(a(),x(v,{key:0},[e(f)?(a(),i(e(N),{key:0})):(a(),i(e($),{key:1,notifications:e(c),onDelete:o[0]||(o[0]=p=>e(t).refresh()),onUpdate:o[1]||(o[1]=p=>e(t).refresh())},null,8,["notifications"]))],64)):C("",!0)]),_:1})}}});export{V as default};
