import{d as _,a as m,c as a,e as d,r as w,o as t,f as n,w as r,k,g as e,R as y,i as g,S as x,T as F,F as h,h as v,U as C}from"./index.52551726.js";const S=_({__name:"Flows",setup($){const u={interval:3e4},s=m(C.getFlows,[{}],u),l=a(()=>{var o;return(o=s.response)!=null?o:[]}),f=a(()=>s.executed&&l.value.length===0),p=a(()=>s.executed);return d("Flows"),(o,c)=>{const i=w("p-layout-default");return t(),n(i,{class:"flows"},{header:r(()=>[k(e(y))]),default:r(()=>[e(p)?(t(),g(h,{key:0},[e(f)?(t(),n(e(x),{key:0})):(t(),n(e(F),{key:1,flows:e(l),onDelete:c[0]||(c[0]=b=>e(s).refresh())},null,8,["flows"]))],64)):v("",!0)]),_:1})}}});export{S as default};
