.flow-run__logs{max-height:100vh}.flow-run__header-meta{display:flex;align-items:center;gap:.5rem}@media (min-width: 1280px){.flow-run__header-meta{display:none}}.flow-run__meta{display:flex;flex-direction:column;align-items:flex-start;gap:.5rem}.flow-run__small-radar{height:250px;width:250px}.flow-run__small-radar-link{display:inline-block;cursor:pointer}
