import{d as _,O as b,c as o,a as f,e as k,r as y,o as m,f as d,w as l,k as c,g as s,aD as g,aE as C,aF as v}from"./index.52551726.js";const x=_({__name:"BlocksCatalog",setup(B){const a=b(null),n=o(()=>a.value?{blockSchemas:{blockCapabilities:{all_:[a.value]}}}:{}),p=f(v.getBlockTypes,[n]),i=o(()=>{var e;return(e=p.response)!=null?e:[]});return k("Blocks Catalog"),(e,t)=>{const r=y("p-layout-default");return m(),d(r,{class:"blocks-catalog"},{header:l(()=>[c(s(g))]),default:l(()=>[c(s(C),{capability:a.value,"onUpdate:capability":t[0]||(t[0]=u=>a.value=u),"block-types":s(i)},null,8,["capability","block-types"])]),_:1})}}});export{x as default};
