import{d as i,u as p,e as m,r as d,o as f,f as k,w as r,k as s,g as t,am as w,an as h,t as u,al as Q,a4 as n}from"./index.52551726.js";const y=i({__name:"WorkQueueCreate",setup(C){const a=p(),c=()=>{a.push(u.workQueues())},_=async o=>{try{const{id:e}=await Q.createWorkQueue(o);n("Work queue has been created","success"),a.push(u.workQueue(e))}catch(e){n("Error occurred while creating new work queue","error"),console.error(e)}};return m("Create Work Queue"),(o,e)=>{const l=d("p-layout-default");return f(),k(l,null,{header:r(()=>[s(t(w))]),default:r(()=>[s(t(h),{action:"Create",onSubmit:_,onCancel:c})]),_:1})}}});export{y as default};
