import{d as U,u as C,c as u,a as R,$ as g,V as x,e as N,N as T,r as c,o as a,f as s,w as n,g as e,a6 as V,h as i,a7 as d,k as p,a8 as F,a9 as P,aa as B,ab as O,Z as Q,t as S,C as W,ac as z}from"./index.52551726.js";import{u as A}from"./useRouteParam.4bc8e74f.js";import{u as G}from"./useToast.e33cfb8b.js";const Z=U({__name:"Deployment",setup(I){const m=A("id"),_=C(),y=G(),f={interval:3e5},v=u(()=>{var l;const o=["Overview","Runs"];return(l=t.value)!=null&&l.deprecated||o.push("Parameters"),W.xl||o.push("Details"),o}),r=R(g.getDeployment,[m.value],f),t=u(()=>r.response);function w(){_.push(S.deployments())}const h=x({deployments:[m.value]}),k=u(()=>t.value?`Deployment: ${t.value.name}`:"Deployment");return N(k),T(t,()=>{var o,l;!((o=t.value)!=null&&o.workQueueName)&&!((l=t.value)!=null&&l.deprecated)&&y(z.info.deploymentMissingWorkQueue,"default",{timeout:!1})}),(o,l)=>{const b=c("p-content"),$=c("p-tabs"),D=c("p-layout-well");return a(),s(D,{class:"deployment"},{header:n(()=>[e(t)?(a(),s(e(V),{key:0,deployment:e(t),onUpdate:e(r).refresh,onDelete:w},null,8,["deployment","onUpdate"])):i("",!0)]),well:n(()=>[e(t)?(a(),s(e(d),{key:0,deployment:e(t),alternate:"",onUpdate:e(r).refresh},null,8,["deployment","onUpdate"])):i("",!0)]),default:n(()=>[e(t)?(a(),s($,{key:0,tabs:e(v)},{overview:n(()=>[p(b,{secondary:""},{default:n(()=>[e(t).deprecated?(a(),s(e(F),{key:0})):e(t).description?(a(),s(e(P),{key:1,description:e(t).description},null,8,["description"])):(a(),s(e(B),{key:2,deployment:e(t)},null,8,["deployment"]))]),_:1})]),parameters:n(()=>[p(e(O),{deployment:e(t)},null,8,["deployment"])]),details:n(()=>[p(e(d),{deployment:e(t),onUpdate:e(r).refresh},null,8,["deployment","onUpdate"])]),runs:n(()=>[p(e(Q),{"flow-run-filter":e(h)},null,8,["flow-run-filter"])]),_:1},8,["tabs"])):i("",!0)]),_:1})}}});export{Z as default};
