import{d as y,a as _,c as a,e as f,r as k,o as n,f as l,w as r,k as $,g as e,a0 as g,i as v,a1 as x,Y as D,F as C,h as b,$ as h}from"./index.52551726.js";const E=y({__name:"Deployments",setup(B){const m={interval:3e4},t=_(h.getDeployments,[{}],m),p=a(()=>{var o;return(o=t.response)!=null?o:[]}),c=a(()=>t.executed&&p.value.length===0),d=a(()=>t.executed);return f("Deployments"),(o,s)=>{const u=k("p-layout-default");return n(),l(u,{class:"deployments"},{header:r(()=>[$(e(g))]),default:r(()=>[e(d)?(n(),v(C,{key:0},[e(c)?(n(),l(e(x),{key:0})):(n(),l(e(D),{key:1,deployments:e(p),onUpdate:s[0]||(s[0]=i=>e(t).refresh()),onDelete:s[1]||(s[1]=i=>e(t).refresh())},null,8,["deployments"]))],64)):b("",!0)]),_:1})}}});export{E as default};
