import{d,a as _,c as u,e as f,r as w,o,f as n,w as c,k as y,g as e,ai as Q,i as g,aj as v,ak as x,F as $,h as C,al as b}from"./index.52551726.js";const B=d({__name:"WorkQueues",setup(h){const l={interval:3e4},s=_(b.getWorkQueues,[{}],l),r=u(()=>{var a;return(a=s.response)!=null?a:[]}),p=u(()=>s.executed&&r.value.length==0),i=u(()=>s.executed);return f("Work Queues"),(a,t)=>{const k=w("p-layout-default");return o(),n(k,{class:"queues"},{header:c(()=>[y(e(Q))]),default:c(()=>[e(i)?(o(),g($,{key:0},[e(p)?(o(),n(e(v),{key:0})):(o(),n(e(x),{key:1,"work-queues":e(r),onUpdate:t[0]||(t[0]=m=>e(s).refresh()),onDelete:t[1]||(t[1]=m=>e(s).refresh())},null,8,["work-queues"]))],64)):C("",!0)]),_:1})}}});export{B as default};
