import{d as w,u as d,a as p,B as R,c as n,e as m,r as v,o as s,f as u,w as t,g as e,E as h,h as F,k as g,G as y,t as k}from"./index.52551726.js";import{u as x}from"./useRouteParam.4bc8e74f.js";const A=w({__name:"FlowRunRadar",setup(C){const r=d(),a=x("id"),l={interval:5e3},c=p(R.getFlowRun,[a.value],l),o=n(()=>c.response);function f(){r.push(k.flowRuns())}const _=n(()=>o.value?`Radar View for Flow Run: ${o.value.name}`:"Radar View for Flow Run");return m(_),(V,B)=>{const i=v("p-layout-full");return s(),u(i,{class:"flow-run-radar"},{header:t(()=>[e(o)?(s(),u(e(h),{key:0,"flow-run":e(o),class:"flow-run-radar__header",onDelete:f},null,8,["flow-run"])):F("",!0)]),default:t(()=>[g(e(y),{"flow-run-id":e(a)},null,8,["flow-run-id"])]),_:1})}}});export{A as default};
